'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var core = require('@rest-hooks/core');
var React = require('react');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefaultLegacy(React);

var REACT_ELEMENT_TYPE;
function _createRawReactElement(type, props, key, children) {
  REACT_ELEMENT_TYPE || (REACT_ELEMENT_TYPE = "function" == typeof Symbol && Symbol["for"] && Symbol["for"]("react.element") || 60103);
  var defaultProps = type && type.defaultProps,
      childrenLength = arguments.length - 3;
  if (props || 0 === childrenLength || (props = {
    children: void 0
  }), 1 === childrenLength) props.children = children;else if (childrenLength > 1) {
    for (var childArray = new Array(childrenLength), i = 0; i < childrenLength; i++) {
      childArray[i] = arguments[i + 3];
    }

    props.children = childArray;
  }
  if (props && defaultProps) for (var propName in defaultProps) {
    void 0 === props[propName] && (props[propName] = defaultProps[propName]);
  } else props || (props = defaultProps || {});
  return {
    $$typeof: REACT_ELEMENT_TYPE,
    type: type,
    key: void 0 === key ? null : "" + key,
    ref: null,
    props: props,
    _owner: null
  };
}

function isNetworkError(error) {
  return Object.prototype.hasOwnProperty.call(error, 'status');
}

/**
 * Handles any networking errors from useResource()
 * @see https://resthooks.io/docs/api/NetworkErrorBoundary
 */
class NetworkErrorBoundary extends React__default["default"].Component {
  constructor(...args) {
    super(...args);
    this.state = {};
  }

  static getDerivedStateFromError(error) {
    if (isNetworkError(error)) {
      return {
        error
      };
    } else {
      throw error;
    }
  }

  render() {
    if (!this.state.error) {
      return this.props.children;
    }

    return /*#__PURE__*/_createRawReactElement(this.props.fallbackComponent, {
      error: this.state.error
    });
  }

}
NetworkErrorBoundary.defaultProps = {
  fallbackComponent: ({
    error
  }) => {
    var _error$response;

    return /*#__PURE__*/_createRawReactElement("div", {}, void 0, error.status, " ", (_error$response = error.response) == null ? void 0 : _error$response.statusText);
  }
};

class BrowserConnectionListener {
  isOnline() {
    if (navigator.onLine !== undefined) {
      return navigator.onLine;
    }

    return true;
  }

  addOnlineListener(handler) {
    addEventListener('online', handler);
  }

  removeOnlineListener(handler) {
    removeEventListener('online', handler);
  }

  addOfflineListener(handler) {
    addEventListener('offline', handler);
  }

  removeOfflineListener(handler) {
    removeEventListener('offline', handler);
  }

}
class AlwaysOnlineConnectionListener {
  isOnline() {
    /* istanbul ignore next */
    return true;
  }

  addOnlineListener() {}

  removeOnlineListener() {}

  addOfflineListener() {}

  removeOfflineListener() {}

}
let DefaultConnectionListener;
/* istanbul ignore if */

if (typeof navigator !== 'undefined' && typeof addEventListener === 'function') {
  DefaultConnectionListener = BrowserConnectionListener;
} else {
  /* istanbul ignore next */
  DefaultConnectionListener = AlwaysOnlineConnectionListener;
}

var DefaultConnectionListener$1 = DefaultConnectionListener;

const {
  FETCH_TYPE
} = core.actionTypes;
/**
 * PollingSubscription keeps a given resource updated by
 * dispatching a fetch at a rate equal to the minimum update
 * interval requested.
 */

class PollingSubscription {
  constructor({
    key,
    schema,
    fetch,
    frequency,
    getState
  }, dispatch, connectionListener) {
    this.frequencyHistogram = new Map();

    this.offlineListener = () => {
      // this clears existing listeners, so no need to clear offline listener
      this.cleanup();
      this.connectionListener.addOnlineListener(this.onlineListener);
    };

    this.onlineListener = () => {
      this.connectionListener.removeOnlineListener(this.onlineListener);
      const now = Date.now();
      this.startId = setTimeout(() => {
        delete this.startId;
        this.update();
        this.run();
      }, Math.max(0, this.lastFetchTime() - now + this.frequency));
      this.connectionListener.addOfflineListener(this.offlineListener);
    };

    if (frequency === undefined) throw new Error('frequency needed for polling subscription');
    this.schema = schema;
    this.fetch = fetch;
    this.frequency = frequency;
    this.key = key;
    this.frequencyHistogram.set(this.frequency, 1);
    this.dispatch = dispatch;
    this.getState = getState;
    this.connectionListener = connectionListener || new DefaultConnectionListener$1(); // Kickstart running since this is initialized after the online notif is sent

    if (this.connectionListener.isOnline()) {
      this.onlineListener();
    } else {
      this.offlineListener();
    }
  }
  /** Subscribe to a frequency */


  add(frequency) {
    if (frequency === undefined) return;

    if (this.frequencyHistogram.has(frequency)) {
      this.frequencyHistogram.set(frequency, this.frequencyHistogram.get(frequency) + 1);
    } else {
      this.frequencyHistogram.set(frequency, 1); // new min so restart service

      if (frequency < this.frequency) {
        this.frequency = frequency;
        this.run();
      }
    }
  }
  /** Unsubscribe from a frequency */


  remove(frequency) {
    if (frequency === undefined) return false;

    if (this.frequencyHistogram.has(frequency)) {
      this.frequencyHistogram.set(frequency, this.frequencyHistogram.get(frequency) - 1);

      if (this.frequencyHistogram.get(frequency) < 1) {
        this.frequencyHistogram.delete(frequency); // nothing subscribed to this anymore...it is invalid

        if (this.frequencyHistogram.size === 0) {
          this.cleanup();
          return true;
        } // this was the min, so find the next size


        if (frequency <= this.frequency) {
          this.frequency = Math.min(...this.frequencyHistogram.keys());
          this.run();
        }
      }
    }
    /* istanbul ignore next */
    else if (process.env.NODE_ENV !== 'production') {
      console.error(`Mismatched remove: ${frequency} is not subscribed for ${this.key}`);
    }

    return false;
  }
  /** Cleanup means clearing out background interval. */


  cleanup() {
    if (this.intervalId) {
      clearInterval(this.intervalId);
      this.intervalId = undefined;
    }

    if (this.lastIntervalId) {
      clearInterval(this.lastIntervalId);
      this.lastIntervalId = undefined;
    }

    if (this.startId) {
      clearTimeout(this.startId);
      delete this.startId;
    }

    this.connectionListener.removeOnlineListener(this.onlineListener);
    this.connectionListener.removeOfflineListener(this.offlineListener);
  }
  /** Trigger request for latest resource */


  update() {
    this.dispatch({
      type: FETCH_TYPE,
      payload: this.fetch,
      meta: {
        schema: this.schema,
        key: this.key,
        type: 'read',
        throttle: true,
        options: {
          dataExpiryLength: this.frequency / 2,
          errorExpiryLength: this.frequency / 10,
          // never break when data already exists
          errorPolicy: () => 'soft'
        },
        createdAt: new Date(),
        resolve: () => {},
        reject: () => {}
      }
    });
  }
  /** What happens when browser goes offline */


  /** Run polling process with current frequency
   *
   * Will clean up old poll interval on next run
   */
  run() {
    if (this.startId) return;
    this.lastIntervalId = this.intervalId;
    this.intervalId = setInterval(() => {
      // since we don't know how long into the last poll it was before resetting
      // we wait til the next fetch to clear old intervals
      if (this.lastIntervalId) {
        clearInterval(this.lastIntervalId);
        this.lastIntervalId = undefined;
      }

      this.update();
    }, this.frequency);
  }
  /** Last fetch time */


  lastFetchTime() {
    var _this$getState$meta$t, _this$getState$meta$t2;

    return (_this$getState$meta$t = (_this$getState$meta$t2 = this.getState().meta[this.key]) == null ? void 0 : _this$getState$meta$t2.date) != null ? _this$getState$meta$t : 0;
  }

}

const {
  SUBSCRIBE_TYPE,
  UNSUBSCRIBE_TYPE
} = core.actionTypes;

/** Handles subscription actions -> fetch or receive actions
 *
 * Constructor takes a SubscriptionConstructable class to control how
 * subscriptions are handled. (e.g., polling, websockets)
 */
class SubscriptionManager {
  constructor(Subscription) {
    this.subscriptions = {};
    this.Subscription = Subscription;

    this.middleware = ({
      dispatch,
      getState
    }) => {
      return next => action => {
        switch (action.type) {
          case SUBSCRIBE_TYPE:
            try {
              this.handleSubscribe(action, dispatch, getState);
            } catch (e) {
              console.error(e);
            }

            return Promise.resolve();

          case UNSUBSCRIBE_TYPE:
            this.handleUnsubscribe(action, dispatch);
            return Promise.resolve();

          default:
            return next(action);
        }
      };
    };
  }
  /** Ensures all subscriptions are cleaned up. */


  cleanup() {
    for (const key in this.subscriptions) {
      this.subscriptions[key].cleanup();
    }
  }
  /** Called when middleware intercepts 'rest-hooks/subscribe' action.
   *
   */


  handleSubscribe(action, dispatch, getState) {
    var _action$meta$options;

    const key = action.meta.key;
    const frequency = (_action$meta$options = action.meta.options) == null ? void 0 : _action$meta$options.pollFrequency;

    if (key in this.subscriptions) {
      this.subscriptions[key].add(frequency);
    } else {
      this.subscriptions[key] = new this.Subscription({
        schema: action.meta.schema,
        fetch: action.meta.fetch,
        frequency,
        key,
        getState
      }, dispatch);
    }
  }
  /** Called when middleware intercepts 'rest-hooks/unsubscribe' action.
   *
   */


  handleUnsubscribe(action, dispatch) {
    var _action$meta$options2;

    const key = action.meta.key;
    const frequency = (_action$meta$options2 = action.meta.options) == null ? void 0 : _action$meta$options2.pollFrequency;
    /* istanbul ignore else */

    if (key in this.subscriptions) {
      const empty = this.subscriptions[key].remove(frequency);

      if (empty) {
        delete this.subscriptions[key];
      }
    } else if (process.env.NODE_ENV !== 'production') {
      console.error(`Mismatched unsubscribe: ${key} is not subscribed`);
    }
  }
  /** Attaches Manager to store
   *
   * Intercepts 'rest-hooks/subscribe'/'rest-hooks/unsubscribe' to register resources that
   * need to be kept up to date.
   *
   * Will possibly dispatch 'rest-hooks/fetch' or 'rest-hooks/receive' to keep resources fresh
   *
   */


  getMiddleware() {
    return this.middleware;
  }

}

/** Integrates with https://github.com/zalmoxisus/redux-devtools-extension
 *
 * Options: https://github.com/zalmoxisus/redux-devtools-extension/blob/master/docs/API/Arguments.md
 */
class DevToolsManager {
  constructor(config = {
    name: `Rest Hooks: ${(() => {
      var _globalThis$document;

      return (_globalThis$document = globalThis.document) == null ? void 0 : _globalThis$document.title;
    })()}`
  }, skipLogging) {
    /* istanbul ignore next */
    this.devTools = typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION__ && window.__REDUX_DEVTOOLS_EXTENSION__.connect(config);
    /* istanbul ignore if */

    /* istanbul ignore next */

    if (process.env.NODE_ENV === 'development' && this.devTools) {
      this.middleware = ({
        getState,
        controller
      }) => {
        const reducer = core.createReducer(controller);
        return next => action => {
          return next(action).then(() => {
            if (skipLogging != null && skipLogging(action)) return;
            const state = getState();
            this.devTools.send(action, state.optimistic.reduce(reducer, state), undefined, 'REST_HOOKS');
          });
        };
      };
    } else {
      this.middleware = () => next => action => next(action);
    }
  }
  /** Called when initial state is ready */


  init(state) {
    /* istanbul ignore if */
    if (this.devTools) this.devTools.init(state);
  }
  /** Ensures all subscriptions are cleaned up. */


  cleanup() {}
  /** Attaches Manager to store
   *
   */


  getMiddleware() {
    return this.middleware;
  }

}

const PromiseifyMiddleware = _ => next => action => {
  next(action);
  return Promise.resolve();
};

var PromiseifyMiddleware$1 = PromiseifyMiddleware;

function ExternalCacheProvider({
  children,
  store,
  selector,
  controller
}) {
  const masterReducer = React.useMemo(() => core.createReducer(controller), [controller]);
  const selectState = React.useCallback(() => {
    const state = selector(store.getState());
    return state.optimistic.reduce(masterReducer, state);
  }, [masterReducer, selector, store]);
  const [state, setState] = React.useState(selectState);
  const isMounted = React.useRef(true);
  React.useEffect(() => () => {
    isMounted.current = false;
  }, []);
  React.useEffect(() => {
    const unsubscribe = store.subscribe(() => {
      if (isMounted.current) setState(selectState());
    });
    return unsubscribe;
  }, [selectState, store]);
  const dispatch = core.usePromisifiedDispatch(store.dispatch, state);
  return /*#__PURE__*/_createRawReactElement(core.DispatchContext.Provider, {
    value: dispatch
  }, void 0, /*#__PURE__*/_createRawReactElement(core.StateContext.Provider, {
    value: state
  }, void 0, /*#__PURE__*/_createRawReactElement(core.ControllerContext.Provider, {
    value: controller
  }, void 0, /*#__PURE__*/_createRawReactElement(core.DenormalizeCacheContext.Provider, {
    value: controller.globalCache
  }, void 0, children))));
}

const mapMiddleware = selector => (...middlewares) => {
  return middlewares.map(middleware => ({
    getState,
    dispatch
  }) => {
    const wrapped = () => selector(getState());

    return middleware({
      getState: wrapped,
      dispatch
    });
  });
};

var mapMiddleware$1 = mapMiddleware;

const CacheProvider = props => core.CacheProvider(props);

CacheProvider.defaultProps = { ...core.CacheProvider.defaultProps,
  managers: [...core.CacheProvider.defaultProps.managers, new SubscriptionManager(PollingSubscription)]
};
/* istanbul ignore next */

if (process.env.NODE_ENV !== 'production') {
  const networkManager = CacheProvider.defaultProps.managers.find(manager => manager instanceof core.NetworkManager);
  CacheProvider.defaultProps.managers.unshift(new DevToolsManager(undefined, networkManager && networkManager.skipLogging.bind(networkManager)));
}

/** Use selector to access part of state */

function useSelectionUnstable(select, params, paramSerializer = p => JSON.stringify(p)) {
  const state = React.useContext(core.StateContext); // TODO: if this is identical to before and render was triggered by state update,
  // we should short-circuit entire rest of render
  // params must be serialized in check
  // eslint-disable-next-line react-hooks/exhaustive-deps

  const selection = React.useMemo(() => params && select(state, params), [// eslint-disable-next-line react-hooks/exhaustive-deps
  params && paramSerializer(params), select, state]);
  return selection;
}

const inferResults = core.__INTERNAL__.inferResults;
const RIC = core.__INTERNAL__.RIC;

var internal = /*#__PURE__*/Object.freeze({
  __proto__: null,
  inferResults: inferResults,
  RIC: RIC,
  initialState: core.initialState,
  StateContext: core.StateContext,
  DispatchContext: core.DispatchContext,
  hasUsableData: core.hasUsableData
});

Object.defineProperty(exports, 'Endpoint', {
  enumerable: true,
  get: function () { return core.Endpoint; }
});
Object.defineProperty(exports, 'Index', {
  enumerable: true,
  get: function () { return core.Index; }
});
Object.defineProperty(exports, 'useCache', {
  enumerable: true,
  get: function () { return core.useCache; }
});
Object.defineProperty(exports, 'useController', {
  enumerable: true,
  get: function () { return core.useController; }
});
Object.defineProperty(exports, 'useDenormalized', {
  enumerable: true,
  get: function () { return core.useDenormalized; }
});
Object.defineProperty(exports, 'useError', {
  enumerable: true,
  get: function () { return core.useError; }
});
Object.defineProperty(exports, 'useFetch', {
  enumerable: true,
  get: function () { return core.useFetch; }
});
Object.defineProperty(exports, 'useFetcher', {
  enumerable: true,
  get: function () { return core.useFetcher; }
});
Object.defineProperty(exports, 'useInvalidator', {
  enumerable: true,
  get: function () { return core.useInvalidator; }
});
Object.defineProperty(exports, 'useMeta', {
  enumerable: true,
  get: function () { return core.useMeta; }
});
Object.defineProperty(exports, 'usePromisifiedDispatch', {
  enumerable: true,
  get: function () { return core.usePromisifiedDispatch; }
});
Object.defineProperty(exports, 'useResetter', {
  enumerable: true,
  get: function () { return core.useResetter; }
});
Object.defineProperty(exports, 'useResource', {
  enumerable: true,
  get: function () { return core.useResource; }
});
Object.defineProperty(exports, 'useRetrieve', {
  enumerable: true,
  get: function () { return core.useRetrieve; }
});
Object.defineProperty(exports, 'useSubscription', {
  enumerable: true,
  get: function () { return core.useSubscription; }
});
Object.defineProperty(exports, 'useSuspense', {
  enumerable: true,
  get: function () { return core.useSuspense; }
});
exports.CacheProvider = CacheProvider;
exports.DefaultConnectionListener = DefaultConnectionListener$1;
exports.DevToolsManager = DevToolsManager;
exports.ExternalCacheProvider = ExternalCacheProvider;
exports.NetworkErrorBoundary = NetworkErrorBoundary;
exports.PollingSubscription = PollingSubscription;
exports.PromiseifyMiddleware = PromiseifyMiddleware$1;
exports.SubscriptionManager = SubscriptionManager;
exports.__INTERNAL__ = internal;
exports.mapMiddleware = mapMiddleware$1;
exports.useSelectionUnstable = useSelectionUnstable;
